% :- type koord      ---> int-int.
% :- type racspontok == list(koord).
% :- type kitoltes   == list(int).

:- module(lepeget_keret, [
			  lepeget_be/2,	% lepeget_be(file::in, racspontok::out) 
			  lepeget_ki/3,	% lepeget_ki(file::in, racspontok::in,
					%            kitoltes::in)
			  megold/2,	% megold(file::in, file::in)
			  stopper/2	% stopper(file::in, file::in)
			]
	 ).

:- use_module(library(lists)).

%-------------- Bejaratok ------------------------------------

% lepeget_be(+File, -Racspontok):
% A File-bol beolvashato egy feladat, a kitoltendo racspontok listaja
% Racspontok.
lepeget_be(File, Racspontok) :-
	open_file(File, read, Stream),
	read_in(Stream, Characters),
	close(Stream),
	parse_data(Characters, Racspontok).

% lepeget_ki(+File, +Racspontok, +Kitoltes):
% Kiirja a File allomanyba azt a megoldast, amelyet ugy kapunk, hogy a
% Racspontok altal leirt racspontokba irva a Kitoltes altal felsorolt
% szamokat.
lepeget_ki(File, Racspontok, Kitoltes) :-
	correct_solution(Racspontok, Kitoltes),
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Racspontok, Kitoltes),
		     (set_output(OOut),close(Stream))).

% megold(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat kiirja a
% FileOut allomanyba.
megold(FileIn, FileOut) :-
	lepeget_be(FileIn, Racspontok),
	ensure_loaded(user:lepeget),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Racspontok),
		     (set_output(OOut),close(Stream))).

% stopper(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper(FileIn, FileOut) :-
	statistics(runtime, [T0|_]),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Feladvny: ~a, id: ~3d s~n', [FileIn,T]).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.  Ha C = -1,
% akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Coords):
% Chars fuzerbol kielemezheto egy Coords koordinata-lista.
parse_data(Chars, Coords) :-
	coordinates(Coords0, Chars, Rest), !,
	correct_syntax(Rest),
	sort(Coords0, Coords).
parse_data(_, _) :-
	warning('Bad format of input file.',[]),
	fail.

%-------------- Ellenorzesek ----------------------------------

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :-
	Rest = [_|_], !, 
	chars_of_type(printing, L, Rest, _),
	warning('non-expected character in: "~s"', [L]),
	fail.
correct_syntax(_).

% correct_solution(+Racspontok, Kitoltes):
% Kitoltes ugyanolyan hosszu, mint Racspontok.
correct_solution(Racspontok, Kitoltes) :-
	length(Racspontok, LR),
	length(Kitoltes, LK),
	(   LR > LK
	->  warning('A megoldslista tl rvid!', []), fail
	;   LR < LK
	->  warning('A megoldslista tl hossz!', []), fail
	;   true
	).

%-------------- DCG elemzes ----------------------------------

% Kielemezheto egy koordinata-lista
coordinates(Coords) -->
	chars_of_type(horiz_sp, _),
	coordinate(Coord), !,
	{Coords = [Coord|Coords1]},
	coordinates(Coords1).
coordinates([]) -->
	chars_of_type(layout, _).

% Kielemezheto egy koordinata
coordinate(Row-Col) -->
	[0'(],
	posint_number(Row),
	[0',],
	posint_number(Col),
	[0')].

% posint_number(Intn) -->
% Kielemezheto egy (tizes szamrendszerbeli) Int pozitiv egesz szam.
posint_number(Int) -->
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% layout_char(Dir, C):
% C egy Dir (horiz_sp|vert_sp) iranyu nem lathato karakter.
layout_char(horiz_sp, 0' ).
layout_char(horiz_sp, 0'\t).
layout_char(vert_sp,  0'\n).
layout_char(vert_sp,  0'\r).
layout_char(vert_sp,  0'\f).
layout_char(vert_sp,  0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	layout_char(_, C).
char_type(printing, C) :-
	\+ layout_char(_, C).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ layout_char(vert_sp, C).
char_type(Dir, C) :-
	layout_char(Dir, C).

%-------------- Kiiratas -------------------------------------

% print_solutions(+Racspontok):
% a Racspontok altal megadott feladat osszes megoldasat kiirja az aktualis
% kimenetre.
print_solutions(Racspontok) :-
	user:lepeget(Racspontok, Kitoltes),
	correct_solution(Racspontok, Kitoltes),
	pretty_print(Racspontok, Kitoltes),
	fail.
print_solutions(_).

% pretty_print(+Racspontok, +Kitoltes)
pretty_print(Racspontok, Kitoltes) :-
	gridsize(Racspontok, RowNum, ColNum),
	make_table(RowNum, ColNum, Table),
	fill_table(Table, Racspontok, Kitoltes),
	print_column_width(Kitoltes, ColWidth),
	print_table(Table, ColWidth).

% gridsize(+Coordinates, RowNum, ColNum)
gridsize(Coords, RowNum, ColNum) :-
	findall(R, member(R-_, Coords), Rs),
	max_list(Rs, RowNum),
	findall(C, member(_-C, Coords), Cs),
	max_list(Cs, ColNum).

% make_table(+RowNum, +ColNum, -Table)
make_table(RowNum, ColNum, T) :-
	length(T, RowNum),
	make_rows(T, ColNum).

% make_rows(+Rows, +RowLen)
make_rows([H|T], L) :-
	length(H, L),
	make_rows(T, L).
make_rows([], _).

% fill_table(+Table, +Coords, -Fill)
fill_table(_, [], []).
fill_table(Table, [C|Cs], [N|Ns]) :-
	mark_at(Table, C, N),
	fill_table(Table, Cs, Ns).

% mark_at(+Table, +Coord, -Number)
mark_at(T, R-C, M) :-
	nth(R, T, Row),
	nth(C, Row, M).

% print_column_width(+Fill, -ColWidth)
print_column_width(Fill, ColWidth) :-
	max_list(Fill, N),
	number_codes(N, NC),
	length(NC, ColWidth).

% print_table(+Table, +ColWidth)
print_table([], _) :- nl.
print_table([R|Rs], ColWidth) :-
	print_row(R, ColWidth),
	print_table(Rs, ColWidth).

% print_row(+Row, +ColWidth)
print_row([], _) :- nl.
print_row([X|Xs], ColWidth) :-
	(   var(X)
	->  format('~|~`-t~*+ ', [ColWidth])
	;   format('~|~t~p~*+ ', [X, ColWidth])
	),
	print_row(Xs, ColWidth).

%-------------- Seged-eljarasok ------------------------------

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).
